#include "clSemiLinkedList.h"

/**
 * Kernel to initialize an empty linked list.
 */
__kernel void
clSemiLinkedList_init(clSemiLinkedList __global *sll) {
	sll->head = NULL;
}

/**
 * Retrieve the next node in the list, from the given node.
 * Wraps to the first node in the list if the current node is marked for deletion (with the least significant bit == 1).
 */
clSemiLinkedList __global *clSemiLinkedList_next(struct clheap __global *heap, clSemiLinkedList __global *list, clSemiLinkedListNode __global *node) {
	uintptr_t next = node->next;
	if((next & 1)!=0) {
		next = list->head;
	}
	return (clSemiLinkedListNode __global *) next;
}

/**
 * Add the given node to the list.
 */
int clSemiLinkedList_add(clSemiLinkedList __global *list, clSemiLinkedListNode __global *node) {
	volatile uintptr_t __global *cursor = (volatile uintptr_t __global *) &(list->head);
	volatile clSemiLinkedListNode __global *cNow;
	uintptr_t cNode;
	unsigned int ret=0;
	node->next = 0;
	while(1) {
		cNode = atom_cmpxchg(cursor, NULL, (uintptr_t) node);
		mem_fence(CLK_GLOBAL_MEM_FENCE);
		if(cNode==NULL) {
			return 1;
		} else {
			cNow = (clSemiLinkedListNode __global *) cNode;
			cursor = (volatile uintptr_t __global *)&(cNow->next);
		}
	}
	return ret;
}

/**
 * Kernel to add keys to the list in parallel.
 */
__kernel
void clSemiLinkedList_addKey(struct clheap __global *heap, clSemiLinkedList __global *list, int __global *keys) {
	int global_x =get_global_id(0);
	clSemiLinkedListNode __global *node = (clSemiLinkedListNode __global *)malloc(heap, sizeof(clSemiLinkedListNode));
	node->key = keys[global_x];
	clSemiLinkedList_add(list, node);
}

/**
 * Find the first occurance in the list, with the specified key.
 */
clSemiLinkedListNode __global *clSemiLinkedList_get(struct clheap __global *heap, clSemiLinkedList __global *list, int key) {
	clSemiLinkedListNode __global *cNow = (clSemiLinkedListNode __global *)(list->head);
	while((cNow != (clSemiLinkedListNode __global *)0) && ((cNow->key) != (key))) {
		cNow = clSemiLinkedList_next(heap, list, cNow);
		/* (clSemiLinkedListNode __global *)(cNow-> next); */
	}
	return cNow;
}

/**
 * Delete the given key from the list.
 */
int clSemiLinkedList_delete(struct clheap __global *heap, clSemiLinkedList __global *list, clSemiLinkedListNode __global *node) {
	volatile uintptr_t __global *cursor = (volatile uintptr_t __global *) &(list->head);
	volatile clSemiLinkedListNode __global *cNow = (volatile clSemiLinkedListNode __global *)(list->head);
	uintptr_t cNode;
	uintptr_t tail = node->next;
	unsigned int ret=0;
	if((tail &1) == 0) {
		/* mark node as in deletion */
		node->next = (node->next)+1;
		while(cNow != (clSemiLinkedListNode __global *)0) {
			cNode = atom_cmpxchg(cursor, (uintptr_t)node, (tail));
			mem_fence(CLK_GLOBAL_MEM_FENCE);
			if(cNode == (uintptr_t)node) {
				free(heap, (uintptr_t)node);
				return 1;
			} else {
				if((cNode & 1) ==0) {
					/* Pointer is clean; Proceed to next */
					cNow = (clSemiLinkedListNode __global *) cNode;
					cursor = (volatile uintptr_t __global *)&(cNow->next);
				} else {
					/* Pointer is dirty; Start from beginning of list. */
					cursor = (volatile uintptr_t __global *) &(list->head);
				}
			}
		}
		/* failed. Restore next field for node */
		node->next = tail;
	}
	return ret;
}

/**
 * Kernel to concurrently delete items with given keys from the list.
 */
__kernel
void clSemiLinkedList_deleteKey(struct clheap __global *heap, clSemiLinkedList __global *list, int __global *keys) {
	int global_x =get_global_id(0);
	clSemiLinkedListNode __global *node = clSemiLinkedList_get(heap, list, keys[global_x]);
	clSemiLinkedList_delete(heap, list, node);
}

/**
 * Kernel to count the number of elements in the list. Should be called with one work-item.
 */
__kernel
void clSemiLinkedList_count(struct clheap __global *heap, clSemiLinkedList __global *list, int __global *count) {
	int i=0;
	clSemiLinkedListNode __global *cNode = (clSemiLinkedList __global *)(list->head);
	int global_x =get_global_id(0);
	if(global_x == 0) {
		while(cNode != NULL) {
			i++;
			cNode = (clSemiLinkedListNode __global *)(cNode->next);
		}
		count[0] = i;
	}
}

__kernel
void clSemiLinkedList_toarray(clSemiLinkedList __global *list, struct clheap __global *heap, __global int*array, int __global *arraySize) {
	clSemiLinkedListNode __global *node = list->head;
	int i=0;
	while(node!=NULL && i<*arraySize) {
		array[i] = node->key;
		node = (clSemiLinkedListNode __global *) (node->next);
		i++;
	}
}

